/**
 * This class defines data types for rigorous integrator odf the KS equation.
 * Note, that the CAPD library provides a general class that computes Poincare map of an abstract system.
 * Parameters of this class are
 * - Solver - that integrates the system
 * - Section - there are three types, but we use AffineSection
 *
 * In capdDynSys/pdes directory of the CAPD library we have implemented classes
 * - PdeSolver - it integrates an abstract Pde provided the vector field is defined
 * - PdeAffineSection - class that represents affine Poincare section
 *
 * This directory also contains code specific for the KS equation
 * - OneDimKSSineVectorField - defines vector field for the KS equation (parameter to PdeSolver) and implements all the bound presented in the article.
 */ 
#ifndef _KS_POINCARE_MAP_
#define _KS_POINCARE_MAP_

#include "capd/capdlib.h"
#include "capd/poincare/PoincareMap.hpp"
#include "capd/poincare/TimeMap.hpp"

#include "capd/pdes/OneDimKSSineVectorField.h"
#include "capd/pdes/PdeSolver.h"
#include "capd/pdes/C0DoubletonSetGeometricTail.h"
#include "capd/pdes/PdeAffineSection.h"

using namespace capd;
using namespace std;
using pdes::GeometricSeries;
using pdes::OneDimKSSineVectorField;
using autodiff::Node;

typedef pdes::PdeSolver<GeometricSeries> PdeSolver;
typedef pdes::PdeAbstractSection<GeometricSeries,IMatrix> PdeAbstractSection;
typedef pdes::PdeAffineSection<GeometricSeries,IMatrix> PdeAffineSection;
typedef poincare::PoincareMap<PdeSolver,PdeAbstractSection> PdePoincareMap;

typedef dynset::FactorReorganization<dynset::PartialQRWithPivoting<5> >  PartialQR;
typedef pdes::C0DoubletonSetGeometricTail<PartialQR> C0DoubletonSetGeometricTail;

// ###############################################
/**
 * This procedure realizes vector field of the 
 * D-dimensional Galerkin projection of the Kuramoto-Sivashinsky PDE
 * with periodic and odd boundary conditions. It is used for simulation and precomputing data for rigorous solver.
 * 
 * Then the vector field is given (for k=1,2,....) by
 * 
 * a_k = k^2*(1-p[0]k^2)*a_k
 *  
 *     - k * \sum_{n=1}^{k-1} a_n*a*_{k-n}
 * 
 *     + 2k * \sum_{n=1}^{D-k} a_n * a_{n+k}
 * 
 * In the procedure below the variables are indexed starting from 0.
 * The first convolution can be optimizes as
 * 
 * 2*k \sum_{n=1}^{k/2} a_n*a_{k-n} 
 * + square of the middle point if k is an even number.
 * 
 * Then the two sums can be partialy factorized which significantly 
 * reduces computational cost, wrapping effect, etc. 
 * 
*/ 
void ksVectorField(
    Node _time, 
    Node in[],   // array of input variables
    int D,       // dimension of domain
    Node out[],  // array of result variables
    int ,        // dimension of counterdomain - here ignored since the same as D
    Node p[],    // array of parameters
    int          // number of parameters
  )
{
  Node one(1.);
  Node half(0.5);
  
  for(int k=1;k<=D;++k)
  {
    int n;
    Node temp(0.);
    int s = capd::min(D-k,(k-1)/2);

    for(n=s+1;n<=(k-1)/2;++n)
      temp -= in[n-1]*in[k-n-1];

    for(n=1;n<=s;++n)
      temp += in[n-1]*(in[n+k-1] - in[k-n-1]);

    for(n=s+1;n<=D-k;++n)
      temp += in[n-1]*in[n+k-1];


    if(k%2==0)
      temp -= half*sqr(in[k/2-1]);
    Node k2(k*k);
    out[k-1] = Node(2*k)*temp + (k2*(one-p[0]*k2))*in[k-1];
  }
}

// ###############################################
/**
 * This class is a wrapper for the (nonrigorous) Poincare map from the CAPD library.
 * It implements D-dimensional Galerkin projection of the KS PDE.
 * It is used for simulation and precomputing data for rigorous solver.
 */ 
struct LDKSPoincareMap {
  /// constructor 
  /// @param [in] D - dimension of the projection
  /// @param [in] nu - viscosity
  /// @param [in] order - order of the numerical (Taylor) method
  /// The default Poincare section is set as x[0] = 0
  LDKSPoincareMap(int D, long double nu, int order=10)
    : vectorField(ksVectorField,D,D,1),
      solver(vectorField,order),
      section(LDVector(D),LDVector(D)),
      pm(solver,section/*,poincare::PlusMinus*/)
  {
    vectorField.setParameter(0,nu);
    LDVector n(D);
    n[0] = 1.;
    section.setNormalVector(n);
  }

  /// This routine sets a new affine Poincare section 
  /// which passes trough the point x
  /// and which is a hyperplane orthogonal to the vector field at x.
  /// @param [in]  x - a vector at which the section is set.
  /// @return - normal vector to the section (normalized vector field at x)
  LDVector setOrthogonalSectionAt(LDVector x){
    LDVector n = vectorField(x);
    n.normalize();
    section = LDAffineSection(x,n);
    return n;
  }

  /// This routine computes n-th iteration of the Poincare map, i.e.
  /// P^n(x) is returned
  /// @param [in] x - a point (not need to be on the section)
  /// @param [in] n - positive integer
  /// @return n-th itersection of the trajectory of x with the section
  LDVector operator()(LDVector x, int n=1){
    for(int i=0;i<n;++i)
      x = pm(x);
    return x;
  }

  /// This routine computes n-th iteration of the Poincare map
  /// and its derivative at x. It returns
  /// P^n(x) and DP^n(x)
  /// @param [in] x - a point (not need to be on the section)
  /// @param [in] n - positive integer
  /// @param [out] D - a matrix tha will be set to DP^n(x)
  /// @return n-th itersection of the trajectory of x with the section
  LDVector operator()(const LDVector& x, LDMatrix& D, int n=1){
    LDVector r = x;
    LDMatrix M(x.dimension(),x.dimension());
    D.setToIdentity();
    for(int i=0;i<n;++i){
      r = pm(r,M); 
      D = pm.computeDP(r,M)*D;
    }
    return r;
  }

  /// This routine computes n-th iteration of the Poincare map
  /// and derivative of the flow at x. It returns
  /// P^n(x) and DPhi^n(x)
  /// @param [in] x - a point (not need to be on the section)
  /// @param [in] n - positive integer
  /// @param [out] D - a matrix tha will be set to DPhi^n(x)
  /// @return n-th itersection of the trajectory of x with the section
  LDVector flowDerivative(const LDVector& x, LDMatrix& D, int n=1){
    LDVector r = x;
    LDMatrix M(x.dimension(),x.dimension());
    D.setToIdentity();
    for(int i=0;i<n;++i){
      r = pm(r,M); 
      D = M*D;
    }
    return r;
  }

  LDMap vectorField;
  LDTaylor solver;
  LDAffineSection section;
  LDPoincareMap pm;
};

// ###############################################
/**
 * This class is a wrapper for the (rigorous) Poincare map from the CAPD library.
 * It provides algorithms for computation of Poincare map of KS PDE in given coordinates.
 */ 
struct IKSPoincareMap {
  /// constructor 
  /// @param [in] _M - dimension of the projection
  /// @param [in] nu - viscosity
  /// @param [in] order - order of the numerical method
  /// @param [in] dissEncDim - index of first dissipative coordinate. Used in the algorithm that finds an enclosure.
  /// @param [in] _q - geometric decay of the tail
  /// The default Poincare section is set as x[0] = 0
  IKSPoincareMap(int dissEncDim, int _M, interval nu, int order, double tol, double _q)
    : M(_M),                                               
      vectorField(nu,_M,dissEncDim),
      solver(vectorField,order),
      section(GeometricSeries(_M,0.,_q),GeometricSeries(_M,0.,_q)),
      pm(solver,section),
      q(_q)
  {
    solver.setAbsoluteTolerance(tol);
    solver.setRelativeTolerance(tol);
    GeometricSeries n(M,0.,q);
    n[0] = 1.;
    section.setNormalVector(n);
  }
  
  /// This routine computes n-th iteration of the Poincare map.
  /// Result is represented in the coordinate system Q centered at u, i.e.
  /// Q*(P^n(x+C*r0)-u) is returned
  /// @param [in] x - center of the set
  /// @param [in] B - coordinate system of a set
  /// @param [in] r - main part of the set, which is x+B*r
  /// @param [in] S - constant used to define tail
  /// @param [in] q - geometric decay used to define tail
  /// @param [in] Q - linear coordinate system in target section
  /// @param [in] u - origin of the coordinate system in target section
  /// @param [in] n - positive integer
  /// @return bound for Q*(P^n(x+B*r+tail)-u)
  GeometricSeries operator()(IVector x, IMatrix B, IVector r, IMatrix Q, IVector u, int n, interval& returnTime, double S){
    C0DoubletonSetGeometricTail initialCondition(toSeries(x,S),B,r);
    returnTime=0.;
    /// Note: the number of explicit coordinates in h-set does not need to be the same as
    ///       M = the number of explicit coordinates used by rigorous integrator PdeSolver
    Q = embedMatrix(Q);
    u = embedVector(u);
    /// call algorithm from CAPD
    return pm(initialCondition,toSeries(u,0.),Q,returnTime,n);
  }

  /// set new Poincare section
  void setSection(const IAffineSection& s){   
    this->section.setOrigin(toSeries(s.getOrigin(),0.));
    this->section.setNormalVector(toSeries(s.getNormalVector(),0.));
  }

private:
  /// This method creates an instance of GeometricSeries from an interval vector x and constant S.
  /// The leading coeffiients of GeometricSeries are taken from x.
  /// The tail is set to S*q^{-i}, for i>dimension(x).
  GeometricSeries toSeries(IVector x, double S){
    GeometricSeries result(M,S,q);
    unsigned i;
    for(i=0;i<x.dimension();++i) result[i] = x[i];
    for(;i<M;++i) result[i] = interval(-S,S)/power(interval(q),i);
    return result;
  }

  /// Given matrix _Q, the function extends it to size MxM by adding 1 at the diagonal
  /// Note: the number of explicit coordinates in h-set does not need to be the same as
  ///       M = the number of explicit coordinates used by rigorous integrator PdeSolver
  IMatrix embedMatrix(IMatrix _Q){
    if(_Q.numberOfRows()==M) return _Q;
    IMatrix Q = IMatrix::Identity(M);
    for(unsigned i=1;i<=_Q.numberOfRows();++i)
      for(unsigned j=1;j<=(int)_Q.numberOfColumns();++j)
        Q(i,j) = _Q(i,j);
    return Q;
  }
  
  /// Given vector x, the funciton extends it to dimension M by adding zeroes at the end.
  /// Note: the number of explicit coordinates in h-set does not need to be the same as
  ///       M = the number of explicit coordinates used by rigorous integrator PdeSolver
  IVector embedVector(IVector x){
    if(x.dimension()==M) return x;
    IVector y(M);
    for(unsigned i=0;i<x.dimension();++i) y[i]=x[i];
    return y;
  }

  int M;
  OneDimKSSineVectorField vectorField;
  PdeSolver solver;
  PdeAffineSection section;
  PdePoincareMap pm;
  double q;
};

#endif
